package mpi.eudico.client.annotator.commands;

import mpi.eudico.client.annotator.ElanFrame2;
import mpi.eudico.client.annotator.linkedmedia.LinkedFileDescriptorUtil;
import mpi.eudico.client.annotator.linkedmedia.MediaDescriptorUtil;
import mpi.eudico.client.annotator.player.ElanMediaPlayer;
import mpi.eudico.client.annotator.util.FrameConstants;

import mpi.eudico.server.corpora.clomimpl.abstr.LinkedFileDescriptor;
import mpi.eudico.server.corpora.clomimpl.abstr.MediaDescriptor;
import mpi.eudico.server.corpora.clomimpl.abstr.TranscriptionImpl;

import java.util.HashMap;
import java.util.Map;
import java.util.Vector;


/**
 * A Command to change the set of linked media or secondary files.
 *
 * @author Han Sloetjes
 */
public class ChangeLinkedFilesCommand implements UndoableCommand {
    private String commandName;

    // receiver; the transcription 
    private TranscriptionImpl transcription;
    private Vector oldDescriptors;
    private Vector newDescriptors;

    // a flag for the kind of descriptors that have been passed to the command
    private boolean areMediaDesc = true;
    // a flag to indicate that only the time offsets of linked files changed
    private boolean timeOffsetsChangedOnly = false;
    // contains mapping from media or link url to an offset. Assumes the same file is not linked more than once.
    private Map<String, Long> oldDescOffsetMap;
    private Map<String, Long> newDescOffsetMap;

    /**
     * Creates a new ChangeLinkedFilesCommand instance
     *
     * @param name the name of the command
     */
    public ChangeLinkedFilesCommand(String name) {
        commandName = name;
    }

    /**
     * The undo action. Restores the old values of the linked files.
     */
    public void undo() {
        if ((transcription != null) && (oldDescriptors != null)) {
            if (areMediaDesc) {
            	if (!timeOffsetsChangedOnly) {
            		updateMediaPlayers(transcription, oldDescriptors);
            	} else {
            		updateMediaPlayersOffset(transcription, oldDescOffsetMap);
            	}
            } else {
            	if (!timeOffsetsChangedOnly) {
            		updateLinkedFiles(transcription, oldDescriptors);
            	} else {
            		updateLinkedFilesOffset(transcription, oldDescOffsetMap);
            	}
            }
        }
    }

    /**
     * The redo action.
     */
    public void redo() {
        if ((transcription != null) && (newDescriptors != null)) {
            if (areMediaDesc) {
            	if (!timeOffsetsChangedOnly) {
            		updateMediaPlayers(transcription, newDescriptors);
            	} else {
            		updateMediaPlayersOffset(transcription, newDescOffsetMap);
            	}
            } else {
            	if (!timeOffsetsChangedOnly) {
            		updateLinkedFiles(transcription, newDescriptors);
            	} else {
            		updateLinkedFilesOffset(transcription, newDescOffsetMap);
            	}
            }
        }
    }

    /**
     * <b>Note: </b>it is assumed the types and order of the arguments are
     * correct.
     *
     * @param receiver the Transcription
     * @param arguments the arguments: <ul><li>arg[0] = the vector with the new
     *        media descriptors or secondary linked files descriptors
     *        (Vector)</li> <li>arg[1] = a flag for the type of descriptors:
     *        if <code>true</code> the descriptors are primary, media
     *        descriptors, if <code>false</code> the descriptors are secondary
     *        linked file descriptors</li> </ul>
     */
    public void execute(Object receiver, Object[] arguments) {
        transcription = (TranscriptionImpl) receiver;

        if ((arguments != null) && (arguments.length >= 1)) {
            newDescriptors = (Vector) arguments[0];

            if (arguments.length > 1) {
                areMediaDesc = ((Boolean) arguments[1]).booleanValue();
            }
            if (arguments.length > 2) {
            	timeOffsetsChangedOnly = ((Boolean) arguments[2]).booleanValue();
            }
        }

        if (transcription != null) {
            if (areMediaDesc) {
                oldDescriptors = transcription.getMediaDescriptors();

                if (!timeOffsetsChangedOnly) {
	                // check if there is any difference??
	                updateMediaPlayers(transcription, newDescriptors);
	                // enable / disable the setFrameLength menu
	                ElanFrame2 ef2 = (ElanFrame2) ELANCommandFactory.getRootFrame(transcription);
	                if (ef2 != null) {
	                    ElanMediaPlayer master = ELANCommandFactory.getViewerManager(transcription).getMasterMediaPlayer();
	                    ef2.setMenuEnabled(FrameConstants.FRAME_LENGTH, !master.isFrameRateAutoDetected());
	                    ef2.updateMenu(FrameConstants.MEDIA_PLAYER);
	                    ef2.updateMenu(FrameConstants.WAVE_FORM_VIEWER);
	                }
                } else {
                	oldDescOffsetMap = new HashMap<String, Long>(oldDescriptors.size());                	
                	newDescOffsetMap = new HashMap<String, Long>(newDescriptors.size());
                	
                	MediaDescriptor md;
                	for (int i = 0; i < oldDescriptors.size(); i++) {
                		md = (MediaDescriptor) oldDescriptors.get(i);
                		oldDescOffsetMap.put(md.mediaURL, md.timeOrigin);
                	}
                	for (int i = 0; i < newDescriptors.size(); i++) {
                		md = (MediaDescriptor) newDescriptors.get(i);
                		newDescOffsetMap.put(md.mediaURL, md.timeOrigin);
                	}
                	
                	updateMediaPlayersOffset(transcription, newDescOffsetMap);
                }
            } else {
                oldDescriptors = transcription.getLinkedFileDescriptors();
                if (!timeOffsetsChangedOnly) {
                	updateLinkedFiles(transcription, newDescriptors);
                } else {
                	oldDescOffsetMap = new HashMap<String, Long>(oldDescriptors.size());
                	newDescOffsetMap = new HashMap<String, Long>(newDescriptors.size());
                	LinkedFileDescriptor lfd;
                	
                	for(int i = 0; i < oldDescriptors.size(); i++) {
                		lfd = (LinkedFileDescriptor) oldDescriptors.get(i);
                		oldDescOffsetMap.put(lfd.linkURL, lfd.timeOrigin);
                	}
                	
                	for(int i = 0; i < newDescriptors.size(); i++) {
                		lfd = (LinkedFileDescriptor) newDescriptors.get(i);
                		newDescOffsetMap.put(lfd.linkURL, lfd.timeOrigin);
                	}
                	
                	updateLinkedFilesOffset(transcription, newDescOffsetMap);
                }
            }
        }
    }

    /**
     * Tries to update the mediaplayers in the viewermanager as well as the
     * layoutmanager and finally sets the mediadescriptors in the
     * transcription.
     *
     * @param transcription the Transcription with the old descriptors
     * @param descriptors the new media descriptors
     */
    private void updateMediaPlayers(TranscriptionImpl transcription,
        Vector descriptors) {
        if ((transcription == null) || (descriptors == null)) {
            return;
        }

        MediaDescriptorUtil.updateMediaPlayers(transcription, descriptors);
    }
    
    /**
     * Tries to update the existing mediaplayers and viewers in the viewermanager as well as the
     * layoutmanager and the media descriptors in the transcription.
     *
     * @param transcription the Transcription with the old descriptors
     * @param offsetMapping the new offsets per media url
     */
    private void updateMediaPlayersOffset(TranscriptionImpl transcription,
        Map<String, Long> newMediaOffsets) {
        if ((transcription == null) || (newMediaOffsets == null)) {
            return;
        }

        MediaDescriptorUtil.updateMediaPlayerOffsets(transcription, newMediaOffsets);
    }

    /**
     * Delegates all updating that needs to be done to a utility class.
     *
     * @param transcription the Transcription with the old descriptors
     * @param descriptors the new linked file descriptors
     */
    private void updateLinkedFiles(TranscriptionImpl transcription,
        Vector descriptors) {
        if ((transcription == null) || (descriptors == null)) {
            return;
        }

        LinkedFileDescriptorUtil.updateLinkedFiles(transcription, descriptors);
    }
    
    /**
     * Updates the offsets of the timeseries tracks and of the media descriptors in 
     * the transcription.
     *
     * @param transcription the Transcription 
     * @param offsetMapping the new offsets per linked file
     */
    private void updateLinkedFilesOffset(TranscriptionImpl transcription,
        Map<String, Long> newLinkOffsets) {
        if ((transcription == null) || (newLinkOffsets == null)) {
            return;
        }
        
        LinkedFileDescriptorUtil.updateLinkedFilesOffsets(transcription, newLinkOffsets);
    }

    /**
     * Returns the name of the command.
     *
     * @return the name of the command
     */
    public String getName() {
        return commandName;
    }
}
