package mpi.eudico.client.annotator.commands;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import mpi.eudico.client.annotator.imports.MergeUtil;
import mpi.eudico.client.annotator.imports.UndoableTranscriptionMerger;
import mpi.eudico.client.annotator.util.AnnotationDataRecord;
import mpi.eudico.server.corpora.clom.Tier;
import mpi.eudico.server.corpora.clomimpl.abstr.AbstractAnnotation;
import mpi.eudico.server.corpora.clomimpl.abstr.TierImpl;
import mpi.eudico.server.corpora.clomimpl.abstr.TranscriptionImpl;

public class MergeTranscriptionsByAddingCommand implements UndoableCommand {
	private String name;
    private TranscriptionImpl firstTrans;
    private TranscriptionImpl secondTrans;
    private UndoableTranscriptionMerger merger;
    
    private List<String> tierNames;
    private boolean processSameNameTiers = false;
    private boolean overwrite = false;
    private boolean addLinkedFiles = false;
    
	/**
	 * A command to add the tiers and annotations from one transcription to another.
	 * This command does not create a copy of the first transcription, but adds directly 
	 * to the first transcription. This means that the action must be undoable.
	 * 
	 * @param name the name of the command
	 */
	public MergeTranscriptionsByAddingCommand(String name) {
		super();
		this.name = name;
	}

	@Override
	public void redo() {
		if (merger != null) {
			merger.redo();
		}

	}

	/**
	 * This method delegates the implementation of undo to the UndoableTranscriptionMerger instance
	 * that did the merging.
	 */
	public void undo() {
		if (merger != null) {
			merger.undo();
		}

	}

	/**
	 * The actual action. The arguments are expected to be in the right order.
	 * 
	 * @param receiver the first, receiving transcription
	 * @param arguments the arguments
	 * arguments[0] = the second transcription (TranscriptionImpl)
	 * arguments[1] = a list of names of tiers to merge/import, can be null in which case all tiers are processed
	 * arguments[2] = a flag to indicate whether tiers with the same name 
	 * in both transcriptions need to be processed as well (Boolean, optional, default: false)
	 * arguments[3] = a flag to indicate whether annotations on tiers with the same name 
	 * may be overwritten or not (Boolean, optional, default: false)
	 * arguments[4] = a flag to indicate whether the linked files of the second transcription 
	 * must be added as well (Boolean, optional, default: false)
	 */
	public void execute(Object receiver, Object[] arguments) {
		firstTrans = (TranscriptionImpl) receiver;
		secondTrans = (TranscriptionImpl) arguments[0];
		tierNames = (List<String>) arguments[1];
		
		if (arguments.length > 2) {
			processSameNameTiers = (Boolean) arguments[2];
		}
		if (arguments.length > 3) {
			overwrite = (Boolean) arguments[3];
		}		
		if (arguments.length > 4) {
			addLinkedFiles = (Boolean) arguments[4];
		}
		
		startMerge();
		
	}

	/**
	 * Starts the merging process.
	 */
	private void startMerge() {
		MergeUtil mergeUtil = new MergeUtil();
		// contains tier objects
		ArrayList<Tier> addableTiers = mergeUtil.getAddableTiers(secondTrans, firstTrans, tierNames);
		List<String> addableTierNames = new ArrayList<String>(addableTiers.size());
		
		TierImpl t;
		for (int i = 0; i < addableTiers.size(); i++) {
			t = (TierImpl) addableTiers.get(i);
			addableTierNames.add(t.getName());
		}
		// call methods on the UndoableTranscriptionMerger?
		merger = new UndoableTranscriptionMerger();
		merger.mergeWith(firstTrans, secondTrans, overwrite, addableTierNames);
	}
	
	/**
	 * Returns the name of the command
	 */
	public String getName() {
		return name;
	}

}
