package mpi.eudico.client.annotator.commands;

import mpi.eudico.client.annotator.util.AnnotationDataRecord;
import mpi.eudico.client.annotator.util.MonitoringLogger;

import mpi.eudico.server.corpora.clom.Annotation;
import mpi.eudico.server.corpora.clom.ExternalReference;
import mpi.eudico.server.corpora.clom.Transcription;

import mpi.eudico.server.corpora.clomimpl.abstr.AbstractAnnotation;
import mpi.eudico.server.corpora.clomimpl.abstr.ExternalReferenceImpl;
import mpi.eudico.server.corpora.clomimpl.abstr.ExternalReferenceGroup;
import mpi.eudico.server.corpora.clomimpl.abstr.TierImpl;

import java.util.ArrayList;

/**
 * A command for modifying an annotation value.
 *
 * @author Han Sloetjes
 */
public class ModifyAnnotationCommand implements UndoableCommand {
    private String commandName;
    private Transcription transcription;
    private AnnotationDataRecord annotationRecord;
    private String oldValue;
    private String newValue;
    private Object oldExtRef;
    private ExternalReference newExtRef;
    private String newCveId;

    /**
     * Creates a new ModifyAnnotationCommand instance
     *
     * @param name the name of the command
     */
    public ModifyAnnotationCommand(String name) {
        commandName = name;
    }

    /**
     * The undo action. We can not just use an object reference to the
     * annotation because the annotation might have been deleted and recreated
     * between the  calls to the execute and undo methods. The reference would
     * then be invalid.
     */
    public void undo() {
        if ((annotationRecord != null) && (transcription != null)) {
            TierImpl tier = (TierImpl) transcription.getTierWithId(annotationRecord.getTierName());
            Annotation annotation = tier.getAnnotationAtTime(annotationRecord.getBeginTime());

            // doublecheck to see if we have the right annotation
            if ((annotation != null) &&
                    (annotation.getEndTimeBoundary() == annotationRecord.getEndTime())) {
                annotation.setCVEntryId(annotationRecord.getCvEntryId());                
                ((AbstractAnnotation) annotation).setExtRef(oldExtRef);
                annotation.setValue(oldValue);
                
                if (MonitoringLogger.isInitiated()) {        	
                	MonitoringLogger.getLogger(transcription).log(MonitoringLogger.UNDO, MonitoringLogger.CHANGE_ANNOTATION_VALUE);        				
                }
            }
        }
    }

    /**
     * The redo action.
     *
     * @see #undo
     */
    public void redo() {
        if ((annotationRecord != null) && (transcription != null)) {
            TierImpl tier = (TierImpl) transcription.getTierWithId(annotationRecord.getTierName());
            Annotation annotation = tier.getAnnotationAtTime(annotationRecord.getBeginTime());

            // doublecheck to see if we have the right annotation
            if ((annotation != null) &&
                    (annotation.getEndTimeBoundary() == annotationRecord.getEndTime())) {
                annotation.setCVEntryId(newCveId);
                //((AbstractAnnotation) annotation).setExtRef(newExtRef);
                replaceExternalReference((AbstractAnnotation)annotation, newExtRef);
                annotation.setValue(newValue);

                if (MonitoringLogger.isInitiated()) {        	
                	MonitoringLogger.getLogger(transcription).log(MonitoringLogger.REDO, MonitoringLogger.CHANGE_ANNOTATION_VALUE);        				
                }
            }                
        }
    }

    /**
     * <b>Note: </b>it is assumed the types and order of the arguments are
     * correct.
     *
     * @param receiver the Annotation (AbstractAnnotation)
     * @param arguments the arguments:  <ul><li>arg[0] = the old value of the
     *        annotation (String)</li> <li>arg[1] = the new value of the
     *        annotation (String)</li> <li>arg[2] = the (new) external reference object 
     *        that can be set by a CV entry</li><li>arg[3] = the (new) id of the controlled
     *        vocabulary entry associated with this annotation</li></ul>
     */
    public void execute(Object receiver, Object[] arguments) {
        AbstractAnnotation annotation = (AbstractAnnotation) receiver;

        if (annotation != null) {
            transcription = (Transcription) annotation.getTier().getParent();
        }

        annotationRecord = new AnnotationDataRecord(annotation);

        oldValue = (String) arguments[0];
        newValue = (String) arguments[1];
        
        // if an ext ref argument is passed, remove the current ext ref of the annotation 
        // even if the new ext ref is null
        if (arguments.length >= 3) {
        	ExternalReferenceImpl eri = (ExternalReferenceImpl) arguments[2];// can be null
        	try {
        		if (annotation.getExtRef() instanceof ExternalReference) {
        			oldExtRef = ((ExternalReference) annotation.getExtRef()).clone();
        		}
			} catch (CloneNotSupportedException e) {
				e.printStackTrace();
			}
        	newExtRef = eri;
        	
        	replaceExternalReference(annotation, eri);
        }
        
        if (arguments.length >= 3) {
            newCveId = (String) arguments[3];        	
        } else {
            newCveId = null;        	        	
        }
        annotation.setCVEntryId(newCveId);
        // Set the text value last: it may trigger update events.
        annotation.setValue(newValue);

        if (MonitoringLogger.isInitiated()) {        	
        	MonitoringLogger.getLogger(transcription).log(MonitoringLogger.CHANGE_ANNOTATION_VALUE);        				
        }
    }

    /*
     * This method is also used in class ModifyOrAddDependentAnnotationsCommand.
     */
	public static void replaceExternalReference(AbstractAnnotation annotation, ExternalReference eri) {
		// List all Ext. Refs that are in the third argument 
		ArrayList<ExternalReference> eriList = new ArrayList<ExternalReference>();
		int listIndex = 0;
		eriList.add(eri);
		
		while (listIndex < eriList.size()) {
			if (eriList.get(listIndex) instanceof ExternalReferenceGroup) {
				eriList.addAll(((ExternalReferenceGroup) eriList.get(listIndex)).getAllReferences());
			}
			listIndex++;
		}

		// Remove all Ext. Refs of the types that are in the list
		for (Object er : eriList) {
			if (er instanceof ExternalReferenceImpl && !(er instanceof ExternalReferenceGroup)) {
				String value = annotation.getExtRefValue(((ExternalReferenceImpl) er).getReferenceType());
				annotation.removeExtRef(new ExternalReferenceImpl(value, ((ExternalReferenceImpl) er).getReferenceType()));
				if (((ExternalReferenceImpl) er).getValue() != null) {
		    		annotation.addExtRef((ExternalReferenceImpl) er);
		    	}
			}
		}
	}

    /**
     * Returns the name of the command.
     *
     * @return the name of the command
     */
    public String getName() {
        return commandName;
    }
}
