package mpi.eudico.client.annotator.interlinear.edit;

import java.awt.FontMetrics;
import java.awt.Graphics2D;
import java.awt.Point;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import mpi.eudico.client.annotator.interlinear.IGTTierType;
import mpi.eudico.client.annotator.interlinear.edit.model.IGTAnnotation;
import mpi.eudico.client.annotator.interlinear.edit.model.IGTDataModel;
import mpi.eudico.client.annotator.interlinear.edit.model.IGTNode;
import mpi.eudico.client.annotator.interlinear.edit.model.IGTTier;
import mpi.eudico.client.annotator.interlinear.edit.render.IGTNodeRenderInfo;
import mpi.eudico.client.annotator.interlinear.edit.render.IGTViewerRenderInfo;

/**
 * A class containing methods for producing the layout, calculating sizes, wrapping lines etc.
 * for an interlinear block.
 * 
 * @author Han Sloetjes
 */
public class IGTCalculator {
	/**
	 * Private constructor, only static methods. 
	 */
	private IGTCalculator() {
	}

	/**
	 * Calculates the y positions of tiers and the total height for the tier in case of wrapping.
	 * 
	 * This method is for tiers that are not part of a subdivision group of tiers.
	 * 
	 * @param viewerRenderInfo the object containing rendering information (fonts, margins, sizes)
	 * @param tier the tier to calculate y positions and height
	 * @param yPosStart the y coordinate (of the top) at which the tier starts
	 * 
	 * @return the total height of the tier
	 */
	public static int calculateTierYPosition(IGTViewerRenderInfo viewerRenderInfo, IGTTier tier, int yPosStart) {
		int curHeight = viewerRenderInfo.getHeightForTier(tier.getTierName());
		tier.getRenderInfo().y = yPosStart;
		
		int totHeight = 0;
		totHeight += (tier.getRenderInfo().numLines * curHeight);
		totHeight += ((tier.getRenderInfo().numLines - 1) * viewerRenderInfo.vertLineMargin);
		tier.getRenderInfo().height = curHeight; // or add the margin to the height?
		tier.getRenderInfo().renderHeight = totHeight;
		
		for (int i = 0; i < tier.getAnnotations().size(); i++) { // should be one annotation
			IGTAnnotation igtAnn = tier.getAnnotations().get(i);
			igtAnn.getRenderInfo().y = yPosStart;
			igtAnn.getRenderInfo().height = totHeight;
		}
		
		totHeight += viewerRenderInfo.vertLineMargin;
		return totHeight;
	}
	
	/**
	 * Calculates the y positions of a tier and its child tiers and the total height 
	 * for the group of tiers (including possible wrapping).
	 * 
	 * This method is for tiers that are the root of a subdivision group of tiers.
	 * 
	 * @param viewerRenderInfo the object containing rendering information (fonts, margins, sizes)
	 * @param tier the root tier to calculate y positions and height
	 * @param yPosStart the y coordinate (of the top) at which the tier starts
	 * 
	 * @return the total height of the tier group
	 */
	public static int calculateTierYPositionRecursive(IGTViewerRenderInfo viewerRenderInfo, IGTTier tier, int yPosStart) {
		int curHeight = viewerRenderInfo.getHeightForTier(tier.getTierName());
		tier.getRenderInfo().clearYPositions();
		tier.getRenderInfo().y = yPosStart;
		tier.getRenderInfo().height = curHeight; // or add the margin to the height?
		int totHeight = 0;
		int numBlocks = 1;
		List<Integer> wrapIndices = tier.getRenderInfo().getWrapIndices();
		if (wrapIndices != null) {
			numBlocks = wrapIndices.size() + 1;
		}
		
		IGTTier depTier;
		IGTAnnotation igtAnn;
		IGTNodeRenderInfo annRender;
		int tempY = yPosStart;
		int deptHeight;
		List<IGTTier> descendants = tier.getDescendantTiers();
		
		for (int j = 0; j < numBlocks; j++) {
			tier.getRenderInfo().addYPosition(tempY);
			totHeight += (curHeight + viewerRenderInfo.vertLineMargin);

			int minA = 0;
			int maxA = tier.getAnnotations().size() - 1;
			if (maxA < 0) {// no annotations
				//continue; 
			}
			
			if (j > 0) {
				minA = wrapIndices.get(j - 1);
			} 
			if (j < numBlocks - 1) {
				maxA = wrapIndices.get(j) - 1;
			}
			
			for (int a = minA; a <= maxA; a++) {
				igtAnn = tier.getAnnotations().get(a);
				igtAnn.getRenderInfo().y = tempY;
				igtAnn.getRenderInfo().height = curHeight;
			}
			
			tempY += (curHeight + viewerRenderInfo.vertLineMargin);
			
			int chMinX = 0;
			int chMaxX = 0;
			if (minA >= 0 && minA < tier.getAnnotations().size()) {
				chMinX = tier.getAnnotations().get(minA).getRenderInfo().x;
			}
			if (maxA >= 0 && maxA < tier.getAnnotations().size()) {
				annRender = tier.getAnnotations().get(maxA).getRenderInfo();
				chMaxX = annRender.x + annRender.calcWidth;
			}
			// don't just take children, but all descendants
			for (int k = 0; k < descendants.size(); k++) {
				depTier = descendants.get(k);
				deptHeight = viewerRenderInfo.getHeightForTier(depTier.getTierName());
				totHeight += (deptHeight + viewerRenderInfo.vertLineMargin);
				if (j == 0) {
					depTier.getRenderInfo().clearYPositions();
				}
				depTier.getRenderInfo().height = deptHeight;
				depTier.getRenderInfo().y = tempY;
				depTier.getRenderInfo().addYPosition(tempY);
				
				// update annotations coordinates
				// on the dep. tier
				for (int a = 0; a < depTier.getAnnotations().size(); a++) {
					igtAnn = depTier.getAnnotations().get(a);
					annRender =igtAnn.getRenderInfo();
					if (annRender.x < chMinX) {
						continue;
					}
					if (annRender.x > chMaxX) {
						break;
					}
					
					annRender.y = tempY;
					annRender.height = deptHeight;
				}
				tempY += (deptHeight + viewerRenderInfo.vertLineMargin);
			}
			
			//tempY += curHeight;
		}
					
		totHeight += viewerRenderInfo.vertLineMargin;
		return totHeight;
	}
	
	/**
	 * Calculates the location and size of the annotations of this tier and depending tiers.
	 * Returns the maximal width the (group of) tier(s) occupies.
	 * 
	 * @param g2d the graphics context
	 * @param viewerRenderInfo the object containing rendering information (fonts, margins, sizes)
	 * @param tier the tier to calculate
	 * @param recursive if true the calculations are performed recursively based on the hierarchy
	 * of the annotations
	 * 
	 * @return the total width (advance) of the tier
	 */
	public static int calculateTierAndAnnotationsLAS(Graphics2D g2d, IGTViewerRenderInfo viewerRenderInfo, 
			IGTTier tier, boolean recursive) {
		if (tier == null) {
			return 0;
		}
		
		List<IGTAnnotation> anns = tier.getAnnotations();
		IGTAnnotation ia;
		int curXAdvance = 0;
		int curXStart = 0;
		FontMetrics fm = g2d.getFontMetrics();// TODO get the actual font metrics per tier		
		
		for (int j = 0; j < anns.size(); j++) {
			ia = anns.get(j);
			int rw = fm.stringWidth(ia.getTextValue());
			ia.getRenderInfo().realWidth = rw;
			ia.getRenderInfo().calcWidth = rw;
			ia.getRenderInfo().x = curXStart; //curXAdvance + viewerRenderInfo.whitespaceWidth;
			curXAdvance += rw;
			// iterate over children. Iterate down and up again.
			// calculate the max width of all children, based on the hierarchy
			if (recursive && ia.getChildCount() > 0) {
				int maxChildrenWidth = 0;
				int curXChildStart = curXStart;
				Map<IGTTier, List<IGTNode>> groupedChildren = ia.getChildrenPerTier();
				Iterator<IGTTier> tierIt = groupedChildren.keySet().iterator();
				while (tierIt.hasNext()) {
					int curWidth = IGTCalculator.calculateChildrenLAS(g2d, viewerRenderInfo, groupedChildren.get(tierIt.next()), 
							curXChildStart /*curXAdvance + viewerRenderInfo.whitespaceWidth*/);
					
					if (curWidth > maxChildrenWidth) {
						maxChildrenWidth = curWidth;
					}
				}
				
				if (maxChildrenWidth > ia.getRenderInfo().calcWidth) {
					ia.getRenderInfo().calcWidth = maxChildrenWidth;
				}
				// update last child per group to match the "word" level calculated width				
				IGTCalculator.updateLastChildSize(ia);

				curXAdvance = ia.getRenderInfo().x + ia.getRenderInfo().calcWidth;
				// end of this group of word level annotation + depending annotations
			}
			
			curXStart += (ia.getRenderInfo().calcWidth + viewerRenderInfo.whitespaceWidth);
		}
		// the total advance of this tiers' annotations
		return curXAdvance;
	}
	
	/**
	 * Method to be called recursively for calculating the width of all children of a particular annotation.
	 * This is always recursive (currently).
	 * 
	 * @param g2d the Graphics context
	 * @param viewerRenderInfo the object containing rendering information (fonts, margins, sizes)
	 * @param children the list of children
	 * @param startX the x coordinate for the x location of the first child per tier
	 * 
	 * @return the total width (advance) of the children
	 */
	public static int calculateChildrenLAS(Graphics2D g2d, IGTViewerRenderInfo viewerRenderInfo, List<IGTNode> children, int startX) {
		if (children == null || children.size() == 0) {
			return 0;
		}
		FontMetrics fm = g2d.getFontMetrics();// get font metrics per tier/font
		IGTAnnotation ia;
		int curXAdvance = startX;
		int totalChildWidth = 0;
		
		for (int i = 0; i < children.size(); i++) {
			ia = (IGTAnnotation) children.get(i);
			int width = fm.stringWidth(ia.getTextValue());
			ia.getRenderInfo().x = curXAdvance;

			ia.getRenderInfo().realWidth = width;
			ia.getRenderInfo().calcWidth = width;
			
			if (ia.getChildCount() > 0) {
				int maxChildrenWidth = 0;
				Map<IGTTier, List<IGTNode>> groupedChildren = ia.getChildrenPerTier();
				Iterator<IGTTier> tierIt = groupedChildren.keySet().iterator();
				while (tierIt.hasNext()) {
					int curWidth = IGTCalculator.calculateChildrenLAS(g2d, viewerRenderInfo, groupedChildren.get(tierIt.next()), 
							ia.getRenderInfo().x);
					
					if (curWidth > maxChildrenWidth) {
						maxChildrenWidth = curWidth;
					}
				}
				if (maxChildrenWidth > width) {
					width = maxChildrenWidth;
					ia.getRenderInfo().calcWidth = maxChildrenWidth;
				}
				
				// update the last child on child tiers to match the total width
				updateLastChildSize(ia);
			}
			
			curXAdvance += width;
			totalChildWidth += width;
			if (i < children.size() - 1) {// was i <= children.size() - 1
				curXAdvance += viewerRenderInfo.whitespaceWidth;
				totalChildWidth += viewerRenderInfo.whitespaceWidth;
			}
		}
		
		return totalChildWidth;
	}
	
	/**
	 * Updates the width of the last child of a node, such that it spans the area up
	 * to the width of this node. Recursive.
	 * 
	 * @param lastNode the last child node on a particular depending tier
	 */
	public static void updateLastChildSize(IGTNode currentNode) {
		if (currentNode != null && currentNode.getChildCount() > 0) {
			Map<IGTTier, List<IGTNode>> groupedChildren = currentNode.getChildrenPerTier();
			Iterator<IGTTier> tierIt = groupedChildren.keySet().iterator();
			IGTNode lastNode;
			while (tierIt.hasNext()) {
				List<IGTNode> children = groupedChildren.get(tierIt.next());
				if (children.size() > 0) {// (un)necessary check?
					lastNode = children.get(children.size() - 1);
					// it shouldn't be possible that the child's calc width > the parent's 
					// calc width, so maybe the next "if" test is not needed 
					if (lastNode.getRenderInfo().x + lastNode.getRenderInfo().calcWidth < 
							currentNode.getRenderInfo().x + currentNode.getRenderInfo().calcWidth) {
						lastNode.getRenderInfo().calcWidth = 
								currentNode.getRenderInfo().x + currentNode.getRenderInfo().calcWidth - 
								lastNode.getRenderInfo().x;
						// update the last children of this last node (so recursive)
						IGTCalculator.updateLastChildSize(lastNode);
					}
				}
			}
		}
	}
	
	/**
	 * Calculates the wrapping info for this tier given the width.
	 * 
	 * @param g2d the Graphics context
	 * @param curTier the tier to wrap
	 * @param width the available width
	 */
	public static void calculateWrappingInfo(Graphics2D g2d, IGTTier curTier, int width) {
		if (curTier == null || width == 0) {
			return;
		}
		FontMetrics fm = g2d.getFontMetrics();// TODO get per tier/font
		
		if (curTier.getType() == IGTTierType.SUBDIVISION || curTier.getType() == IGTTierType.WORD_LEVEL_ROOT) {
			IGTCalculator.calculateBlockWrapping(curTier, width);
		} else if (curTier.getType() == IGTTierType.ROOT || curTier.getType() == IGTTierType.ASSOCIATION 
				|| curTier.getType() == IGTTierType.TIME_CODE){ // TODO check if already processed and/or make this test more fine grained
			List<IGTAnnotation> anns = curTier.getAnnotations();
			if (anns.size() > 0) {
				IGTAnnotation ann = curTier.getAnnotations().get(0);
				//calculateLineWrapping(g2d, fm, ann, width);
				IGTCalculator.calculateLineWrappingSimple(g2d, fm, ann, width);
			}
		}
	}
	
	/**
	 * Calculates the wrapping information for a single line of text.
	 * The approach in this method is to just start at the beginning add words to test until a 
	 * wrapping position is found.
	 * 
	 * @param g2d the graphics context
	 * @param fm the font metrics
	 * @param ann the annotation to calculate wrapping for
	 * @param width the maximum width for a line 
	 */
	public static void calculateLineWrappingSimple(Graphics2D g2d, FontMetrics fm, IGTAnnotation ann, int width) {
		if (ann == null || ann.getTextValue() == null || ann.getTextValue().length() == 0) {
			return;
		}
		
		if (ann.getRenderInfo().realWidth > width) {
			ann.getRenderInfo().clearWrappedLines();
			String total = ann.getTextValue().replace("\n", " ");// text value can't be null here
			int[] indices = getSpaceIndices(total);
			
			if (indices.length == 0) {
				ann.getRenderInfo().numLines = 1;
				ann.getIGTTier().getRenderInfo().numLines = 1;
			} else {
				int bIndex = 0;
				int eIndex = 0;
				
				String substring = null;
				String prevSubstring = null;
				int subWidth;
				int numLines = 0;
				
				for (int i = 0; i < indices.length; i++) {
					eIndex = indices[i];
					substring = total.substring(bIndex, eIndex);
					subWidth = fm.stringWidth(substring);
					
					if (subWidth > width) {// crossed the boundary, use previous substring
						if (i == 0) {
							ann.getRenderInfo().addWrappedLine(substring);
							numLines++;
							bIndex = eIndex + 1;
						} else if (i == indices.length - 1) {
							if (prevSubstring != null) {
								ann.getRenderInfo().addWrappedLine(prevSubstring);
								numLines++;
							}
							bIndex = indices[i - 1] + 1;
							substring = total.substring(bIndex);
							subWidth = fm.stringWidth(substring);
							if (subWidth < width) {
								ann.getRenderInfo().addWrappedLine(substring);
								numLines++;
							} else {// add two lines
								ann.getRenderInfo().addWrappedLine(total.substring(bIndex, eIndex));
								numLines++;
								ann.getRenderInfo().addWrappedLine(total.substring(eIndex + 1));
								numLines++;
							}
							break;
						} else {// default
							if (prevSubstring != null) {
								ann.getRenderInfo().addWrappedLine(prevSubstring);
								numLines++;
								bIndex = indices[i - 1] + 1;
							} else {
								ann.getRenderInfo().addWrappedLine(substring);
								numLines++;
								bIndex = eIndex + 1;
							}							
						}
						
					} else {// else continue with the same begin index
						if (i == indices.length - 1) {// last index
							prevSubstring = substring;
							substring = total.substring(bIndex);// try the rest of the string
							subWidth = fm.stringWidth(substring);
							if (subWidth < width) {
								ann.getRenderInfo().addWrappedLine(substring);
								numLines++;
							} else {
								ann.getRenderInfo().addWrappedLine(prevSubstring);
								numLines++;
								ann.getRenderInfo().addWrappedLine(total.substring(eIndex + 1));
								numLines++;
							}
						} else {
							prevSubstring = substring;
							continue;
						}						
					}
					prevSubstring = null;
				}
				
				ann.getRenderInfo().numLines = numLines;
				ann.getIGTTier().getRenderInfo().numLines = numLines;
			}
		} else {
			ann.getRenderInfo().numLines = 1;
			ann.getRenderInfo().clearWrappedLines();
			ann.getIGTTier().getRenderInfo().numLines = 1;
		}
	}
	
	/**
	 * Calculates the annotation indices at which wrapping should occur. The index of the annotation is 
	 * that of the first annotation that should move to a new line.
	 * Individual words are currently not wrapped.
	 * 
	 * @param tier the tier which is parent on the level of subdivision, e.g. on the word level
	 * 
	 * @param width the width of display area
	 */
	public static void calculateBlockWrapping(IGTTier tier, int width) {
		if (tier != null && 
				(tier.getType() == IGTTierType.SUBDIVISION || tier.getType() == IGTTierType.WORD_LEVEL_ROOT)) {
			tier.getRenderInfo().clearWrapIndices();

			int numBlocks = 1;
			IGTAnnotation ann;
			int xShift = 0;
			
			for (int i = 0; i < tier.getAnnotations().size(); i++ ) {
				ann = tier.getAnnotations().get(i);
				int calcW = ann.getRenderInfo().calcWidth;
				
				// add a special check for the last annotation?
				// this block might need to be removed
				if (i == tier.getAnnotations().size() - 1) {
					int realW = ann.getRenderInfo().realWidth;
					List<IGTNode> children = ann.getChildren();
					if (children != null && children.size() > 0) {
						IGTAnnotation ca;
						for (int j = 0; j < children.size(); j++) {
							ca = (IGTAnnotation) children.get(j);
							if (ca.getRenderInfo().realWidth > realW) {
								realW = ca.getRenderInfo().realWidth;
							}
						}
					}
					calcW = realW;
				}
				
				if (ann.getRenderInfo().x + calcW >= (numBlocks * width)) {
					// wrap
					tier.getRenderInfo().addWrapIndex(i);
					numBlocks++;
					xShift = ann.getRenderInfo().x;
				}
				ann.getRenderInfo().x -= xShift;
				// update child annotations
				List<IGTNode> children = ann.getChildren();
				if (children != null && children.size() > 0) {
					for (IGTNode child : children) {
						child.getRenderInfo().x -= xShift;
					}
				}
			}
		}
	}

	/**
	 * Returns an array of indices of whitespace characters in a string.
	 * 
	 * @param s the string
	 * @return the array, if there are no whitespace characters the array is empty
	 */
	public static int[] getSpaceIndices(String s) {
		if (s == null) {
			return new int[]{};
		}
		List<Integer> inds = new ArrayList<Integer>();
		for (int i = 0; i < s.length(); i++) {
			if (Character.isWhitespace(s.charAt(i))) {
				inds.add(i);
			}
		}
		if (inds.size() > 0) {
			int[] indices = new int[inds.size()];
			for (int i = 0; i < indices.length; i++) {
				indices[i] = inds.get(i);
			}
			return indices;
		} else {
			return new int[]{};
		}
	}
	
	/**
	 * Calculates the shortened label for a the tiers in the model.
	 * 
	 * @param g2d the graphics context for the rendering
	 * @param dModel the data model for one block of annotations
	 * @param availWidth the available width for the labels. Margins have to be taken into account
	 */
	public static void calculateShortTierLabels(Graphics2D g2d, IGTDataModel dModel, int availWidth) {
		if (dModel == null || availWidth <= 0) {// the width can have a greater minimum
			return;// silently return or throw exception?
		}
		FontMetrics fm = g2d.getFontMetrics();// TODO get per tier/font
		//int netWidth = availWidth - dModel.getRenderInfo().margin;
		int netWidth = availWidth;// TODO include margins
		
		IGTTier igtTier;
		for (int i = 0; i < dModel.getRowCount(); i++) {
			igtTier = dModel.getRowData(i);
			int totalWidth = fm.stringWidth(igtTier.getTierName());
			if (totalWidth >= availWidth) {
				String shortName = igtTier.getTierName().substring(0, igtTier.getTierName().length() - 1);
				while (fm.stringWidth(shortName) >= netWidth && shortName.length() > 1) {
					shortName = shortName.substring(0, shortName.length() - 1);
				}
				
				dModel.setShortTierNameForIndex(i, shortName);
			} else {
				// set the short width to ?
				dModel.setShortTierNameForIndex(i, igtTier.getTierName());
			}
		}
	}
	
	/**
	 * Updates part of an igt group starting with a node that has to be updated and then
	 * propagating changes to descending annotations and then up the tree again.  
	 * TODO make sure annotations are not shifted out of the view
	 * 
	 * @param igtAnn the annotation from which the update starts
	 * @param calcWidth new calculated width for the node
	 */
	public static void updateLASRecursiveFromNode(IGTAnnotation igtAnn, int calcWidth) {
		if (igtAnn == null) {
			return;
		}
		int widthDiff = calcWidth - igtAnn.getRenderInfo().calcWidth;
		// if widthDiff > 0 the node needs more space, subsequent annotations are 
		// shifted to the right, 
		// TODO the widthDiff < 0, following annotations are shifted left, is not implemented yet.  
		// A check on the real width should be included in that case.
		if (widthDiff <= 0) {
			return;
		}
		igtAnn.getRenderInfo().calcWidth += widthDiff;
		// or move up first??
		updateLastChildSize(igtAnn);
		// find the parent that is on the "word level" tier 
		IGTTier wlTier = null;
		IGTAnnotation curAnn = igtAnn;
		IGTTier curTier = curAnn.getIGTTier();
		if (curTier.isInWordLevelBlock()) {
			IGTTier parentTier;
			while (true) {
				parentTier = curTier.getParentTier();
				if (parentTier == null || !parentTier.isInWordLevelBlock()) {
					wlTier = curTier;
					break;
				} else {
					curTier = parentTier;
				}
			}
		}
		
		if (wlTier != null && wlTier != igtAnn.getIGTTier()) {
			IGTAnnotation parentAnn = wlTier.getAnnotationAtPoint(new Point(curAnn.getRenderInfo().x, 
					wlTier.getRenderInfo().y));
			if (parentAnn != null) {// it shouldn't be, otherwise the way to get to the ancestor should be changed 
				parentAnn.getRenderInfo().calcWidth += widthDiff;
				curAnn = parentAnn;
			}
		}
		// shift next annotations
		IGTAnnotation nextAnn = curAnn.getIGTTier().getNextAnnotation(curAnn);
		while (nextAnn != null) {
			nextAnn.getRenderInfo().x += widthDiff;
			// move children?
			List<IGTNode> children = nextAnn.getChildren();
			if (children != null) {
				for (IGTNode node : children) {
					node.getRenderInfo().x += widthDiff;
				}
			}
			
			nextAnn = curAnn.getIGTTier().getNextAnnotation(nextAnn);
		}

		// update top level tiers and 
		IGTTier topTier = igtAnn.getIGTTier();
		while (topTier != null) {
			if (topTier.getParentTier() != null) {
				topTier = topTier.getParentTier();
			} else {
				break;
			}
		}
		
		if (topTier != null && topTier != igtAnn.getIGTTier()) {
			nextAnn = topTier.getAnnotations().get(0);
			if (nextAnn != null) {
				nextAnn.getRenderInfo().calcWidth += widthDiff;
				updateLastChildSize(nextAnn);
			}
		}
		
	}
	
}
