package mpi.eudico.client.annotator.interlinear.edit.model;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import mpi.eudico.client.annotator.interlinear.IGTTierType;
import mpi.eudico.client.annotator.interlinear.edit.event.IGTDataModelEvent;
import mpi.eudico.client.annotator.interlinear.edit.event.IGTDataModelListener;
import mpi.eudico.client.annotator.interlinear.edit.event.ModelEventType;
import mpi.eudico.client.annotator.interlinear.edit.render.IGTBlockRenderInfo;
import mpi.eudico.server.corpora.clomimpl.abstr.AbstractAnnotation;
import mpi.eudico.server.corpora.clomimpl.abstr.AlignableAnnotation;
import mpi.eudico.server.corpora.clomimpl.abstr.TierImpl;
import mpi.eudico.server.corpora.clomimpl.type.Constraint;
import mpi.eudico.util.TimeFormatter;

/**
 * Implementation of an igt data model based on a group of annotations,
 * with a single top level annotation as the root.
 * 
 * @author Han Sloetjes
 *
 */
public class IGTDefaultModel extends IGTAbstractDataModel {
	
	/**
	 * Constructor.
	 * @param rootAnnotation the top level alignable annotation
	 */
	public IGTDefaultModel(AlignableAnnotation rootAnnotation) {
		rowData = new ArrayList<IGTTier>();
		rowHeader = new IGTRowHeader();
		renderInfo = new IGTBlockRenderInfo();
		listeners = new ArrayList<IGTDataModelListener>(4);
		fillTable(rootAnnotation);
	}

	private void fillTable(AlignableAnnotation rootAnnotation) {
		IGTAnnotation ia = new IGTAnnotation(rootAnnotation);
		IGTTier igtTier = new IGTTier(((TierImpl) rootAnnotation.getTier()).getName(), IGTTierType.ROOT);
		ia.type = IGTTierType.ROOT;
		rowData.add(igtTier);
		rowHeader.addHeader(((TierImpl) rootAnnotation.getTier()).getName());
		igtTier.addAnnotation(ia);
		// add place holder tiers and/or fill "empty" positions with empty place holder nodes
		addDependingTiersNodes(igtTier, (TierImpl) rootAnnotation.getTier());
		
		// add children
		addChildrenToNode(rootAnnotation.getParentListeners(), ia);
		
		// add nodes for speaker, time info, silence duration etc?
		addSpecialTiers(ia);
	}
	
	/**
	 * Recursively creates and adds tier nodes for child tiers and adds them to the model. 
	 * Currently all tiers are just added, no support yet for a selection of tiers.
	 * 
	 * @param parentTierNode the parent node
	 * @param parentTier the parent tier
	 */
	private void addDependingTiersNodes(IGTTier parentTierNode, TierImpl parentTier) {
		List depTiers = parentTier.getChildTiers();
		
		if (depTiers != null && depTiers.size() > 0) {
			TierImpl dt;
			IGTTierType type = IGTTierType.ASSOCIATION;
			boolean inWLBlock = false;
			IGTTier igtTier = null;
			
			for (int i = 0; i < depTiers.size(); i++) {
				dt = (TierImpl) depTiers.get(i);
				//System.out.println("Parent: " + parentTier.getName() + " - Child " + i + ": "+ dt.getName());
				Constraint c = dt.getLinguisticType().getConstraints();
				if (c.getStereoType() == Constraint.SYMBOLIC_ASSOCIATION) {
					if (parentTierNode.getType() == IGTTierType.ROOT) {
						type = IGTTierType.FIRST_LEVEL_ASSOCIATION;
						inWLBlock = false;
					} else {
						type = IGTTierType.DEEPER_LEVEL_ASSOCIATION;
						inWLBlock = parentTierNode.isInWordLevelBlock();
					}
				} else if (c.getStereoType() == Constraint.SYMBOLIC_SUBDIVISION || 
						c.getStereoType() == Constraint.TIME_SUBDIVISION ||
						c.getStereoType() == Constraint.INCLUDED_IN) {
					if (parentTierNode.getType() == IGTTierType.ROOT || 
							parentTierNode.getType() == IGTTierType.FIRST_LEVEL_ASSOCIATION) {
						type = IGTTierType.WORD_LEVEL_ROOT;//?? can't assume this, could be phrase level as well
						inWLBlock = true;
					} else {
						type = IGTTierType.SUBDIVISION;	
						inWLBlock = true;
					}
				}
				
				igtTier = new IGTTier(dt.getName(), type);
				igtTier.setParentTier(parentTierNode);
				parentTierNode.addChildTier(igtTier);
				igtTier.setInWordLevelBlock(inWLBlock);
				rowData.add(igtTier);
				rowHeader.addHeader(dt.getName());
				// recursive
				addDependingTiersNodes(igtTier, dt);
			}
		}
	}
	
	/**
	 * Adds nodes for child annotations of a parent annotation to the appropriate tier node while constructing
	 * tree relations between the annotations themselves.
	 * 
	 * @param childAnnos the child annotations
	 * @param parentNode the node of the parent annotation
	 */
	private void addChildrenToNode(List childAnnos, IGTAnnotation parentNode) {
		if (childAnnos == null || parentNode == null) {
			return;			
		}
		IGTTier parentTier = parentNode.getIGTTier();
		Map<String, List<AbstractAnnotation>> amap = new HashMap<String, List<AbstractAnnotation>>();
		AbstractAnnotation aa;
		List<AbstractAnnotation> group;
		String tierName;
		
		for (int i = 0; i < childAnnos.size(); i++) {
			aa = (AbstractAnnotation) childAnnos.get(i);
			tierName = aa.getTier().getName();
			group = amap.get(tierName);
			if (group == null) {
				group = new ArrayList<AbstractAnnotation>();
				amap.put(tierName, group);
			}
			group.add(aa);
		}
		aa = null;
		group = null;
		tierName = null;
		
		Iterator<String> tierIt = amap.keySet().iterator();
		IGTTierType type = IGTTierType.ASSOCIATION;
		boolean inWLBlock = false;
		IGTTier igtTier = null;
		
		while (tierIt.hasNext()) {
			tierName = tierIt.next();
			int tierIndex = getRowIndexForTier(tierName);
			if (tierIndex == -1) {
				igtTier = null;
			} else {
				igtTier = rowData.get(tierIndex);
			}
			
			group = amap.get(tierName);
			if (group.size() > 1) {
				Collections.sort(group, new AAComparator());
			}
			
			for (int i = 0; i < group.size(); i++) {
				aa = group.get(i);
				if (i == 0) {
					Constraint c = ((TierImpl) aa.getTier()).getLinguisticType().getConstraints();
					if (c.getStereoType() == Constraint.SYMBOLIC_ASSOCIATION) {
						if (parentTier.getType() == IGTTierType.ROOT) {
							type = IGTTierType.FIRST_LEVEL_ASSOCIATION;
							inWLBlock = false;
						} else {
							type = IGTTierType.DEEPER_LEVEL_ASSOCIATION;
							inWLBlock = parentTier.isInWordLevelBlock();
						}
					} else if (c.getStereoType() == Constraint.SYMBOLIC_SUBDIVISION || 
							c.getStereoType() == Constraint.TIME_SUBDIVISION ||
							c.getStereoType() == Constraint.INCLUDED_IN) {
						if (parentTier.getType() == IGTTierType.ROOT || 
								parentTier.getType() == IGTTierType.FIRST_LEVEL_ASSOCIATION) {
							type = IGTTierType.WORD_LEVEL_ROOT;//?? can't assume this, could be phrase level as well
							inWLBlock = true;
						} else {
							type = IGTTierType.SUBDIVISION;	
							inWLBlock = true;
						}
					}
				}
				
				IGTAnnotation ia = new IGTAnnotation(aa);
				if (igtTier == null) { // this shouldn't occur if the tiers have already been added
					igtTier = new IGTTier(tierName, type);
					igtTier.setInWordLevelBlock(inWLBlock);
					parentTier.addChildTier(igtTier);
					igtTier.setParentTier(parentTier);
					rowData.add(igtTier);
					rowHeader.addHeader(tierName);
				}
				igtTier.addAnnotation(ia);
				parentNode.addChild(ia);
				
				List nextChildren = aa.getParentListeners();
				if (nextChildren.size() > 0) {
					addChildrenToNode(nextChildren, ia);
				}
			}
		}
	}
	
	/**
	 * Adds "placeholder tiers" for time and speaker information. 
	 * 
	 * @param ia the root annotation
	 */
	private void addSpecialTiers(IGTAnnotation ia) {
		long bt = ia.getAnnotation().getBeginTimeBoundary();
		long et = ia.getAnnotation().getEndTimeBoundary();
		IGTAnnotation timeAnnNode = new IGTAnnotation(TimeFormatter.toString(bt) + " - " + TimeFormatter.toString(et));
		ia.addChild(timeAnnNode);
		IGTTier tiTierNode = new IGTTier("TC", IGTTierType.TIME_CODE);// localize??
		tiTierNode.addAnnotation(timeAnnNode);
		rowData.add(tiTierNode);
		rowHeader.addHeader(tiTierNode.getTierName());
		ia.getIGTTier().addChildTier(tiTierNode);
		tiTierNode.setParentTier(ia.getIGTTier());
		
		String speaker = ((TierImpl) ia.getAnnotation().getTier()).getParticipant();
		if (speaker == null) {
			speaker = "";// or Undefined? Unknown
		}
		IGTAnnotation speakAnnNode = new IGTAnnotation("");
		ia.addChild(speakAnnNode);
		IGTTier speakTierNode = new IGTTier("Speaker", IGTTierType.SPEAKER_LABEL);// localize??
		speakTierNode.addAnnotation(speakAnnNode);
		rowData.add(speakTierNode);
		rowHeader.addHeader(speakTierNode.getTierName());
		ia.getIGTTier().addChildTier(speakTierNode);
		speakTierNode.setParentTier(ia.getIGTTier());
		
		// add silence duration tier
	}

	/**
	 * Finds the root tier and if that tier has an annotation (one IGTAnnotation containing an annotation), 
	 * its begin time is returned.
	 * 
	 * @see mpi.eudico.client.annotator.interlinear.edit.model.IGTDataModel#getBeginTime()
	 */
	@Override
	public long getBeginTime() {
		IGTTier root = getRootRow();
		if (root != null) {
			if (root.getAnnotations().size() > 0) {// should be 1
				IGTAnnotation rootAnn = root.getAnnotations().get(0);
				if (rootAnn.getAnnotation() != null) {
					return rootAnn.getAnnotation().getBeginTimeBoundary();
				}
			}
		}
		
		return -1;
	}

	/**
	 * Finds the root tier and if that tier has an annotation (one IGTAnnotation containing an annotation), 
	 * its end time is returned.
	 * 
	 * @see mpi.eudico.client.annotator.interlinear.edit.model.IGTDataModel#getEndTime()
	 */
	@Override
	public long getEndTime() {
		IGTTier root = getRootRow();
		if (root != null) {
			if (root.getAnnotations().size() > 0) {// should be 1
				IGTAnnotation rootAnn = root.getAnnotations().get(0);
				if (rootAnn.getAnnotation() != null) {
					return rootAnn.getAnnotation().getEndTimeBoundary();
				}
			}
		}
		
		return -1;
	}

	// methods for updating the model as far as annotations are concerned, reflect the type of events in ACMEditEvent
	
	/**
	 * Adds an IGT annotation for the specified Annotation to the model, if it can find the right position, the right
	 * annotation group to add it to.
	 * 
	 * @param annotation the annotation to add to the model
	 */
	public void annotationAdded(AbstractAnnotation annotation) {
		if (annotation != null) {
			IGTTier chTier = getRowDataForTier(annotation.getTier().getName());
			IGTTier parTier = chTier.getParentTier();// check null??
			List<IGTAnnotation> parAnns = parTier.getAnnotations();
			IGTAnnotation parAnn;
			for (int j = 0; j < parAnns.size(); j++) {
				parAnn = parAnns.get(j);
				// TODO there are several ways to check which annotation is the parent annotation 
				// one is via times. Another one is via ParentListeners
				if (parAnn.getAnnotation().getBeginTimeBoundary() <= annotation.getBeginTimeBoundary() && 
						parAnn.getAnnotation().getEndTimeBoundary() >= annotation.getEndTimeBoundary()) {
					IGTAnnotation nextIgtAnn = new IGTAnnotation(annotation);
					boolean inserted = false;
					
					// check if there are already child annotations 
					Map<IGTTier, List<IGTNode>> childMap = parAnn.getChildrenPerTier();
					if (childMap != null) {
						List<IGTNode> childList = childMap.get(chTier);
						if (childList != null) {
							//find position
							IGTAnnotation child;
							for (int k = 0; k < childList.size(); k++) {
								child = (IGTAnnotation) childList.get(k);
								if (child.getAnnotation().getBeginTimeBoundary() >= annotation.getEndTimeBoundary()) {
									//  insert before this child, assume that existing annotations times have been updated
									inserted = parAnn.insertChild(nextIgtAnn, k);
								} else if (k == childList.size() - 1) {
									// after the last existing annotation
									inserted = parAnn.addChild(nextIgtAnn);
								}
							}
						} else {// no children for this parent annotation on this tier yet
							inserted = parAnn.addChild(nextIgtAnn);
						}
					} else {// no children on any tier yet, add
						inserted = parAnn.addChild(nextIgtAnn);
					}
					
					if (inserted) {
						nextIgtAnn.setIGTTier(chTier);
						IGTAnnotation child;
						if (chTier.getAnnotations().size() == 0) {
							chTier.addAnnotation(nextIgtAnn);
						} else {
							for (int n = 0; n < chTier.getAnnotations().size(); n++) {
								child = chTier.getAnnotations().get(n);
								if (child.getAnnotation().getBeginTimeBoundary() >= annotation.getEndTimeBoundary()) {
									chTier.insertAnnotation(nextIgtAnn, n);
									break;
								} 
								else if (n == chTier.getAnnotations().size() - 1) {
									chTier.addAnnotation(nextIgtAnn);// situation of annotation after?
									break;
								}
							}
						}
						// generate and post event
						IGTDataModelEvent event = new IGTDataModelEvent(this, chTier, nextIgtAnn, ModelEventType.ADD);
						postEvent(event);
					}
					break;
				}
			}
		}
	}

	/**
	 * Adds an annotation that has been inserted after an existing annotation.
	 * 
	 * @param annotation the annotation to add to the model
	 * @param afterThis the annotation the new annotation has been inserted after
	 */
	public void annotationAddedAfter(AbstractAnnotation annotation,
			AbstractAnnotation afterThis) {
		if (annotation != null && afterThis != null) {
			IGTTier chTier = getRowDataForTier(annotation.getTier().getName());
			int numAnns = chTier.getAnnotations().size();
			IGTAnnotation afterAnn;
			
			for (int i = 0; i < numAnns; i++) {
				afterAnn = chTier.getAnnotations().get(i);
				if (afterAnn.getAnnotation() == afterThis) {
					
					IGTAnnotation nextIgtAnn = new IGTAnnotation(annotation);
					boolean inserted = false;
					
					IGTAnnotation parAnn = (IGTAnnotation) afterAnn.getParent();
					int afterIndex = parAnn.getIndex(afterAnn);
					//assert(afterIndex != -1);
					if (afterIndex > -1) {
						if (afterIndex < parAnn.getChildCount()) {
							inserted = parAnn.insertChild(nextIgtAnn, afterIndex + 1);
						} else {
							inserted = parAnn.addChild(nextIgtAnn);
						}
					}
					
					if (inserted) {
						if (i < numAnns - 2) {
							chTier.insertAnnotation(nextIgtAnn, i + 1);
						} else {
							chTier.addAnnotation(nextIgtAnn);
						}
						nextIgtAnn.setIGTTier(chTier);
						
						// generate and post event
						IGTDataModelEvent event = new IGTDataModelEvent(this, chTier, nextIgtAnn, ModelEventType.ADD);
						postEvent(event);
						
						break;
					}
				}
			}
		}		
	}

	/**
	 * Adds an annotation that has been inserted before an existing annotation
	 * 
	 * @param annotation the annotation to add to the model
	 * @param beforeThis the annotation the new annotation has been inserted before
	 */
	public void annotationAddedBefore(AbstractAnnotation annotation,
			AbstractAnnotation beforeThis) {
		if (annotation != null && beforeThis != null) {
			IGTTier chTier = getRowDataForTier(annotation.getTier().getName());
			int numAnns = chTier.getAnnotations().size();
			IGTAnnotation beforeAnn;
			
			for (int i = 0; i < numAnns; i++) {
				beforeAnn = chTier.getAnnotations().get(i);
				if (beforeAnn.getAnnotation() == beforeThis) {
					
					IGTAnnotation nextIgtAnn = new IGTAnnotation(annotation);
					boolean inserted = false;
					
					IGTAnnotation parAnn = (IGTAnnotation) beforeAnn.getParent();
					int beforeIndex = parAnn.getIndex(beforeAnn);
					//assert(beforeIndex != -1);
					if (beforeIndex > -1) {
						inserted = parAnn.insertChild(nextIgtAnn, beforeIndex);
					}
					
					if (inserted) {
						chTier.insertAnnotation(nextIgtAnn, i);
						nextIgtAnn.setIGTTier(chTier);
						// generate and post event
						IGTDataModelEvent event = new IGTDataModelEvent(this, chTier, nextIgtAnn, ModelEventType.ADD);
						postEvent(event);
						
						break;
					}
				}
			}
		}		
	}

	/**
	 * Changes the text value of an existing annotation. 
	 * 
	 * @param annotation the changed annotation
	 */
	public void annotationValueChanged(AbstractAnnotation annotation) {
		if (annotation != null) {
			IGTTier chTier = getRowDataForTier(annotation.getTier().getName());
			int numAnns = chTier.getAnnotations().size();
			IGTAnnotation igtAnn;
			
			for (int i = 0; i < numAnns; i++) {
				igtAnn = chTier.getAnnotations().get(i);
				if (igtAnn.getAnnotation() == annotation) {
					// the annotation is indeed in this model, recalculate the rendering
					// generate and post event
					IGTDataModelEvent event = new IGTDataModelEvent(this, chTier, igtAnn, ModelEventType.CHANGE);
					postEvent(event);
					
					break;
				}
			}
		}		
	}

	/**
	 * Removes the IGT annotation for the specified annotation.
	 * 
	 * @param annotation the annotation to remove
	 */
	public void annotationRemoved(AbstractAnnotation annotation) {
		if (annotation != null) {
			IGTTier chTier = getRowDataForTier(annotation.getTier().getName());
			int numAnns = chTier.getAnnotations().size();
			IGTAnnotation igtAnn;
			
			for (int i = 0; i < numAnns; i++) {
				igtAnn = chTier.getAnnotations().get(i);
				if (igtAnn.getAnnotation() == annotation) {
					// the annotation is indeed in this model, disconnect and recalculate the rendering
					
					removeChildren(igtAnn);
					if (igtAnn.getParent() != null) {
						igtAnn.getParent().removeChild(igtAnn);
					}
					chTier.removeAnnotation(igtAnn);
					
					// generate and post event
					IGTDataModelEvent event = new IGTDataModelEvent(this, chTier, igtAnn, ModelEventType.REMOVE);
					postEvent(event);
					
					break;
				}
			}
		}
	}
	
	/**
	 * Removes all children of the specified annotation/node. Recursively.
	 * 
	 * @param igtAnnotation the IGT annotation of which to delete all children
	 */
	private void removeChildren(IGTAnnotation igtAnnotation) {
		if (igtAnnotation != null) {

			int numChildren = igtAnnotation.getChildCount();
			IGTAnnotation child;
			for (int i = numChildren - 1; i >= 0; i--) {
				child = (IGTAnnotation) igtAnnotation.getChildren().get(i);
				igtAnnotation.removeChild(child);
				child.getIGTTier().removeAnnotation(child);
				removeChildren(child);
			}
		}
	}

	// methods to notify listeners
	protected void notifyRowAdded() {
		IGTDataModelEvent event = new IGTDataModelEvent(this);
		for (int i = 0; i < listeners.size(); i++) {
			listeners.get(i).dataModelChanged(event);
		}
	}
	
	/**
	 * Mini implementation of a comparator. Compares annotations on the same tier, 
	 * assumes that they different begin times. 
	 * 
	 * @author Han Sloetjes
	 */
	private class AAComparator implements Comparator<AbstractAnnotation> {

		/**
		 * Compares only time info.
		 */
		public int compare(AbstractAnnotation arg0, AbstractAnnotation arg1) {
			if (arg0.getBeginTimeBoundary() < arg1.getBeginTimeBoundary()) {
				return -1;
			} else if (arg0.getBeginTimeBoundary() > arg1.getBeginTimeBoundary()) {
				return 1;
			}
			// here could compare end times
			return 0;
		}
		
	}

	
}
