package mpi.eudico.client.annotator.interlinear.edit.model;

import java.util.List;

import mpi.eudico.client.annotator.interlinear.IGTTierType;
import nl.mpi.lexan.annotyzers.helpers.Position;
import nl.mpi.lexan.annotyzers.helpers.Suggestion;
import nl.mpi.lexan.annotyzers.helpers.SuggestionSet;


/**
 * A data model for suggestion sets.
 * Mainly used to store the data structure of a set of suggestions for rendering purposes.
 * 
 * @author Han Sloetjes
 */
public class IGTSuggestionModel extends IGTAbstractDataModel {
	private SuggestionSet suggestionSet;
	private long beginTime = Long.MAX_VALUE;
	private long endTime = -1;

	/**
	 * Constructor with a Suggestion Set as parameter. 
	 * 
	 * @param suggestionSet a set of suggestions for (possibly hierarchically grouped) annotations 
	 */
	public IGTSuggestionModel(SuggestionSet suggestionSet) {
		super();
		this.suggestionSet = suggestionSet;
		fillModel();
		// hier... calculate sizes?
	}
	
	/**
	 * Returns the SuggestionSet which is the basis for the model.
	 * 
	 * @return the suggestion set
	 */
	public SuggestionSet getSuggestionSet() {
		return suggestionSet;
	}
	
	/**
	 * Extracts information from the suggestion set and builds the model.
	 * Currently adds an empty, place holder tier as the root,
	 * and treat the first level suggestions as "subdivision" (word level root).
	 */
	private void fillModel() {
		if (suggestionSet != null) {
			Suggestion suggestion = null;
			Position pos;
			// add invisible root tier for all other tiers
			IGTTier rootTier = new IGTTier("ROOT", IGTTierType.ROOT);
			rowData.add(rootTier);
			rowHeader.addHeader("ROOT");
			IGTAnnotation rootAnn = new IGTAnnotation("ROOT");
			rootAnn.type = IGTTierType.ROOT;
			rootTier.addAnnotation(rootAnn);
			
			for (int i = 0; i < suggestionSet.suggestions.size(); i++) {
				suggestion = suggestionSet.suggestions.get(i);
				pos = suggestion.position;
				
				IGTTier wlRootTier = getRowDataForTier(pos.tierId);
				
				if (wlRootTier == null) {
					// special treatment of the first suggestion, determines the "root"
					wlRootTier = new IGTTier(pos.tierId, IGTTierType.WORD_LEVEL_ROOT);
					wlRootTier.setInWordLevelBlock(true);
					rowData.add(wlRootTier);
					rowHeader.addHeader(pos.tierId);
					rootTier.addChildTier(wlRootTier);
					wlRootTier.setParentTier(rootTier);
				}
				
				IGTAnnotation igtAnn = new IGTAnnotation(suggestion.content);
				igtAnn.type = IGTTierType.WORD_LEVEL_ROOT;
				wlRootTier.addAnnotation(igtAnn);
				rootAnn.addChild(igtAnn);
				
				if (pos.beginTime < beginTime) {
					beginTime = pos.beginTime;
				}
				if (pos.endTime > endTime) {
					endTime = pos.endTime;
				}
				// add children
				addChildNodes(wlRootTier, igtAnn, suggestion.children);
			}
		}
	}
	
	/**
	 * Adds the child suggestions to the model
	 *  
	 * @param parentTier the parent tier
	 * @param parentAnn the parent annotation
	 * @param children the list of child suggestions
	 */
	private void addChildNodes(IGTTier parentTier, IGTAnnotation parentAnn, List<Suggestion> children) {
		if (children != null) {
			Position pos;
			for (Suggestion sug : children) {
				pos = sug.position;
				
				IGTTier childTier = getRowDataForTier(pos.tierId);
				
				if (childTier == null) {
					childTier = new IGTTier(pos.tierId, IGTTierType.SUBDIVISION);
					childTier.setInWordLevelBlock(true);// set all tiers to be in a word level block for now
					rowData.add(childTier);
					rowHeader.addHeader(pos.tierId);
					parentTier.addChildTier(childTier);
					childTier.setParentTier(parentTier);
				}
				
				IGTAnnotation igtAnn = new IGTAnnotation(sug.content);
				igtAnn.type = IGTTierType.SUBDIVISION;
				childTier.addAnnotation(igtAnn);
				parentAnn.addChild(igtAnn);
				
				if (pos.beginTime < beginTime) {
					beginTime = pos.beginTime;
				}
				if (pos.endTime > endTime) {
					endTime = pos.endTime;
				}
				
				// add children
				addChildNodes(childTier, igtAnn, sug.children);
			}
		}
	}

	@Override
	public long getBeginTime() {
		if (beginTime != Long.MAX_VALUE) {
			return beginTime;
		}
		
		return 0;
	}

	@Override
	public long getEndTime() {
		if (endTime > -1) {
			return endTime;
		}
		return 0;
	}

	
}
