package mpi.eudico.client.annotator.prefs;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.lang.ref.WeakReference;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.xml.sax.Attributes;
import org.xml.sax.SAXException;
import org.xml.sax.XMLReader;
import org.xml.sax.helpers.DefaultHandler;
import org.xml.sax.helpers.XMLReaderFactory;

import mpi.eudico.client.annotator.Constants;
import mpi.eudico.client.annotator.FrameManager;
import mpi.eudico.client.annotator.ShutdownListener;
import mpi.eudico.util.IoUtil;
import mpi.eudico.util.multilangcv.LangInfo;

public class RecentLanguages implements ShutdownListener {

	public static final long maxAge = 366L * 24L * 60L * 60L;	// one year (including a leap day) in seconds.
	public static final String privatePreferencesFile = "RecentLanguages.xml";
	
	private static class RecentLangInfo extends LangInfo {
		private long used;
		
		public RecentLangInfo(LangInfo other) {
			super(other);
			used = System.currentTimeMillis() / 1000;
		}

		public RecentLangInfo(String id, String longId, String label) {
			super(id, longId, label);
			used = System.currentTimeMillis() / 1000;
		}
		
		public long getUsed() {
			return used;
		}		
	}
	
	private List<RecentLangInfo> recentLanguages;
	private List<WeakReference<RecentLanguageListener>> listeners;
	
	private static RecentLanguages instance;
	
	/**
	 * Get the singleton RecentLanguages instance.
	 * To be called from the GUI thread: this is not thread-safe.
	 * @return
	 */
	public static RecentLanguages getInstance() {
		if (instance != null) {
			return instance;
		}
		instance = new RecentLanguages();
		return instance;
	}

	/**
	 * Nobody else should be able to call the constructor directly.
	 */
	private RecentLanguages() {
	}
	
	/**
	 * Access function for use by the RecentLanguagesMenu and RecentLanguagesBox.
	 * By making the list unmodifiable we can also claim a super type as element type.
	 * @return
	 */
	public List<LangInfo> getRecentLanguages() {
		syncWithPrivatePreferences();
		return Collections.<LangInfo>unmodifiableList(recentLanguages);
	}

	/**
	 * Keep the recent list up-to-date.
	 * Languages not used for a year get forgotten.
	 * This method is typically called from BasicControlledVocabulary, when anything 
	 * regarding a LangInfo changes.
	 * 
	 * @param li the changed LangInfo
	 */
	public void addRecentLanguage(LangInfo newLI) {
		syncWithPrivatePreferences();
				
		// See if we already have this language in our list.
		int size = recentLanguages.size();
		for (int i = 0; i < size; i++) {
			LangInfo li = recentLanguages.get(i);
			if (li == newLI || li.getLongId().equals(newLI.getLongId())) {
				// Language already in list... maybe other fields were edited.
				// Update the use time in any case.
				recentLanguages.set(i, new RecentLangInfo(newLI));
				notifyListeners(i, newLI);
				return;
			}				
		}
		// Need to add it. Use a free index number if possible.
		int index = recentLanguages.size();
		recentLanguages.add(new RecentLangInfo(newLI));

		notifyListeners(index, newLI);
	}

	/**
	 * Convert the Preferences values describing the list of recently used languages
	 * to a private form which is easier to process.
	 */
	
	private void syncWithPrivatePreferences() {
		if (recentLanguages != null) {
			// We've already done the work before...
			return;
		}
		recentLanguages = new ArrayList<RecentLangInfo>();
	
		XMLReader reader;
		
        try {
            reader = XMLReaderFactory.createXMLReader(
                "org.apache.xerces.parsers.SAXParser");
            reader.setFeature("http://xml.org/sax/features/namespaces", false);
            reader.setFeature("http://xml.org/sax/features/validation", false);
            reader.setFeature("http://apache.org/xml/features/validation/schema", false);
            reader.setFeature("http://apache.org/xml/features/validation/dynamic", false);
            RecentLanguagesHandler rlh = new RecentLanguagesHandler();
            reader.setContentHandler(rlh);
            reader.setErrorHandler(rlh);
            // This works to make sure the schema isn't fetched from the web but from here:
            //reader.setEntityResolver(new MyResolver()); // see http://www.saxproject.org/apidoc/org/xml/sax/EntityResolver.html
            String fileName = Constants.ELAN_DATA_DIR + File.separator + privatePreferencesFile;
            reader.parse(fileName);
        } catch (SAXException e) {
            e.printStackTrace();
        } catch (FileNotFoundException e) {
        	// If the file is not there, that's not a problem.
        } catch (IOException e) {
            e.printStackTrace();
		}
        
        FrameManager.getInstance().addWindowCloseListener(this);
	}
	
	/**
	 * Save the recent languages to a private "preferences" file.
	 * Create DOM tree and then save it.
	 */
	private void savePrivatePreferences() {
	    DocumentBuilderFactory dbf;
	    DocumentBuilder db;

        dbf = DocumentBuilderFactory.newInstance();
        try {
			db = dbf.newDocumentBuilder();
			
			Document doc = db.newDocument();
			Element root = doc.createElement("RECENT_LANGUAGES");
			doc.appendChild(root);
			
			for (RecentLangInfo li : recentLanguages) {
				Element l = doc.createElement("LANGUAGE");
				l.setAttribute("LANG_ID", li.getId());
				l.setAttribute("LANG_DEF", li.getLongId());
				l.setAttribute("LANG_LABEL", li.getLabel());
				l.setAttribute("used", String.valueOf(li.getUsed()));
				
				root.appendChild(l);
			}
			
			// DOM tree finished; now write a file.
			
            String fileName = Constants.ELAN_DATA_DIR + File.separator + privatePreferencesFile;
			IoUtil.writeEncodedFile("UTF-8", fileName, doc.getDocumentElement());
			
		} catch (ParserConfigurationException e) {
			e.printStackTrace();
		} catch (Exception e) {
			e.printStackTrace();
		}
	}
	
	/*
	 * Handle listeners to our changes.
	 * Typically the RecentLanguagesMenuItem.
	 * Use WeakReference<RecentLanguageListener> so that we don't keep
	 * listener objects alive if they never could call
	 * removeRecentLanguageListener().
	 */

	public void addRecentLanguageListener(
			RecentLanguageListener listener) {
		if (listeners == null) {
			listeners = new ArrayList<WeakReference<RecentLanguageListener>>(1);
		}
		listeners.add(new WeakReference<RecentLanguageListener>(listener));
	}

	public void removeRecentLanguageListener(
			RecentLanguageListener listener) {
		if (listeners != null) {

			Iterator<WeakReference<RecentLanguageListener>> iter = listeners.iterator();
			
			while (iter.hasNext()) {
				WeakReference<RecentLanguageListener> ref = iter.next();
				if (ref.get() == listener || ref.get() == null) {
					iter.remove();
				}
			}
		}
	}

	private void notifyListeners(int freeIndex, LangInfo newLI) {
		if (listeners != null) {
			Iterator<WeakReference<RecentLanguageListener>> iter = listeners.iterator();
			
			while (iter.hasNext()) {
				WeakReference<RecentLanguageListener> ref = iter.next();
				RecentLanguageListener listener = ref.get();
				if (listener == null) {
					iter.remove();
				} else {
					listener.recentLanguageAdded(freeIndex, newLI);
				}
			}
		}
	}
	
	/**
	 * Handler for parsing the list of recent languages
	 * 
	 * @author olasei
	 */
	private class RecentLanguagesHandler extends DefaultHandler {
		private long oldestAge;
		
		public RecentLanguagesHandler() {
			oldestAge = System.currentTimeMillis() / 1000 - maxAge;
		}

        @Override
        public void startElement(String nameSpaceURI, String name, String rawName,
                Attributes attrs) throws SAXException {

        	if ("LANGUAGE".equals(rawName)) {
        		String id = attrs.getValue("LANG_ID");
        		String def = attrs.getValue("LANG_DEF");
        		String label = attrs.getValue("LANG_LABEL");
        		String used = attrs.getValue("used");
        		
        		RecentLangInfo rli = new RecentLangInfo(id, def, label);
        		try {
        			rli.used = Long.valueOf(used);
        		} catch (NumberFormatException nfe) {
        			rli.used = 0;
        		}
        		if (rli.used > oldestAge) {
        			recentLanguages.add(rli);
        		}
        	}
        }
	}

	/**
	 * Called when a window, or Elan, is closing.
	 */
	@Override
	public void somethingIsClosing(ShutdownListener.Event e) {
		if (e.getType() == Event.ELAN_EXITS_EARLY) {
			savePrivatePreferences();
		}
	}
}