package mpi.eudico.client.annotator.webserviceclient.weblicht;

import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Insets;
import java.io.IOException;

import javax.swing.JLabel;
import javax.swing.JProgressBar;
import javax.swing.border.EmptyBorder;

import org.xml.sax.SAXException;

import mpi.eudico.client.annotator.commands.ELANCommandFactory;
import mpi.eudico.client.annotator.commands.MergeTranscriptionsByAddingCommand;
import mpi.eudico.client.annotator.gui.multistep.MultiStepPane;
import mpi.eudico.client.annotator.gui.multistep.StepPane;
import mpi.eudico.client.annotator.util.ClientLogger;
import mpi.eudico.server.corpora.clomimpl.abstr.TranscriptionImpl;
import mpi.eudico.webserviceclient.weblicht.TCFtoTranscription;
import mpi.eudico.webserviceclient.weblicht.WebLichtWsClient;
/**
 * The final step in uploading text to weblicht and converting the result to a transcription.
 * 
 * @author Han Sloetjes
 *
 */
public class WebLichtStep4 extends StepPane {
	private JLabel progressLabel;
	private JProgressBar progressBar;
	
	// some fields for the settings resulting from the previous steps
	// necessary because the interaction takes place in a separate thread
	private String inputText;
	private int sentenceDuration;
	private boolean includeTokenize;
	private boolean includePosTags;
	private boolean includeLemmas;
	
	private boolean cancelled = false;
	
	/**
	 * Constructor
	 * @param multiPane
	 */
	public WebLichtStep4(MultiStepPane multiPane) {
		super(multiPane);
		initComponents();
	}

	
	@Override
	protected void initComponents() {
		setLayout(new GridBagLayout());
		setBorder(new EmptyBorder(5, 10, 5, 10));
		
		progressLabel = new JLabel("Processing text... please wait");
		progressBar = new JProgressBar(0, 100);
		
		Insets insets = new Insets(2, 0, 2, 0);
		GridBagConstraints gbc = new GridBagConstraints();
		gbc.anchor = GridBagConstraints.WEST;
		gbc.insets = insets;
		gbc.fill = GridBagConstraints.HORIZONTAL;
		gbc.weightx = 1.0;
		
		add(progressLabel, gbc);
		gbc.gridy = 1;
		add(progressBar, gbc);
	}


	@Override
	public String getStepTitle() {
		return "Uploading to WebLicht";
	}

	@Override
	public void enterStepForward() {
		// everything has been set here. Start processing
		multiPane.setButtonEnabled(MultiStepPane.ALL_BUTTONS, false);
		// get all info from the properties, upload text in a separate thread
		inputText = (String) multiPane.getStepProperty("InputText");
		sentenceDuration = (Integer) multiPane.getStepProperty("SentenceDuration");
		includeTokenize = (Boolean) multiPane.getStepProperty("IncludeTokenize");
		includePosTags = (Boolean) multiPane.getStepProperty("IncludePOS");
		includeLemmas = (Boolean) multiPane.getStepProperty("IncludeLemma");
		
		multiPane.setButtonEnabled(MultiStepPane.CANCEL_BUTTON, true);
		
		new WLThread().start();
	}

	
	@Override
	public void cancelled() {
		cancelled = true;
	}


	@Override
	public void finished() {
		cancelled = true;
	}

	private void closeInternal() {
		multiPane.close();
	}

	/**
	 * A thread that performs the calling of the web services.
	 * 
	 * @author Han Sloetjes
	 *
	 */
	private class WLThread extends Thread {

		/**
		 * Starts uploading text and possibly calls additional services.
		 */
		public void run() {
			progressLabel.setText("Uploading text to WebLicht...");
			if (cancelled) {
				return;
			}
			WebLichtWsClient wsClient = new WebLichtWsClient();
			String tcfToParse = null;
			// hier configure
			String tcf = wsClient.convertPlainText(inputText);
			// immediately upload the contents to a sentence splitter
			if (tcf != null) {
				if (cancelled) {
					return;
				}
				//System.out.println(tcf);
				tcfToParse = tcf;
				
				if (includeTokenize) {
					progressLabel.setText("Uploading sentences for tokenization...");
					// this should be detected dynamically?
					//String sentenceTokenUrl = "service-opennlp-1_5/tcf/detect-sentences/tokenize";
					String sentenceTokenUrl = "service-opennlp/annotate/tok-sentences";
					String tcf2 = wsClient.callWithTCF(sentenceTokenUrl, tcf);
					
					if (tcf2 != null) {
						//System.out.println(tcf2);
						if (cancelled) {
							return;
						}
						tcfToParse = tcf2;
						String tcf3 = null;
						if (includePosTags) {
							progressLabel.setText("Recieved tokens, uploading for Part-Of-Speech");
							// hardcoded
							String posTagUrl = "service-opennlp/annotate/postag";
						
							tcf3 = wsClient.callWithTCF(posTagUrl, tcf2);
							if (tcf3 != null) {
								//System.out.println(tcf3);
								if (cancelled) {
									return;
								}
								
								tcfToParse = tcf3;
								progressLabel.setText("Received Part-Of-Speech tags");
							} else {
								progressLabel.setText("An error occurred: could not get the Part-Of-Speech tags");
							}
						}
						String tcf4 = null;
						
						if (includeLemmas) {// lemmatizer creates pos tags as well?
							if (cancelled) {
								return;
							}
							progressLabel.setText("Uploading results for lemma's");
							
							String lemmaURL = "service-lemmatizer/annotate/lemma-pos";// hier CHECK
							if (tcf3 != null) {
								tcf4 = wsClient.callWithTCF(lemmaURL, tcf3);							
							} else {
								tcf4 = wsClient.callWithTCF(lemmaURL, tcf2);	
							}
							
							if (tcf4 != null) {
								//System.out.println(tcf4);
								tcfToParse = tcf4;
								progressLabel.setText("Received Lemma's");
							} else {
								progressLabel.setText("An error occurred: could not get the Lemma's");
							}
						}
					} else {
						progressLabel.setText("Unable to get tokens, creating tiers(s)...");
					}
				}
				
				if (tcfToParse != null) {
					if (cancelled) {
						return;
					}
					progressLabel.setText("Creating and adding tiers(s)...");
					TCFtoTranscription tctt = new TCFtoTranscription();
					try {
						TranscriptionImpl trans2 = tctt.createTranscription(tcfToParse);
						TranscriptionImpl trans = (TranscriptionImpl) multiPane.getStepProperty("transcription");
							
						MergeTranscriptionsByAddingCommand command = (MergeTranscriptionsByAddingCommand) 
								ELANCommandFactory.createCommand(trans, ELANCommandFactory.MERGE_TRANSCRIPTIONS_UNDOABLE);
						command.execute(trans, new Object[]{trans2, null, Boolean.TRUE, Boolean.TRUE, Boolean.FALSE});
							
					} catch (SAXException sax) {
						ClientLogger.LOG.warning("Error parsing tcf file: " + sax.getMessage());
					} catch (IOException ioe) {
						ClientLogger.LOG.warning("Error parsing tcf file: " + ioe.getMessage());
					}
					// popup a message?
					
				}
			} else {
				progressLabel.setText("Error while uploading the text, no results created");
			}
			// ready show message? and close the window
			closeInternal();
		} 
		
	}
}
